@@ -80,6 +80,22 @@ class permittedKwargs:
         return wrapped
 
 
+class permittedMethodKwargs:
+
+    def __init__(self, permitted):
+        self.permitted = permitted
+
+    def __call__(self, f):
+        @wraps(f)
+        def wrapped(obj, args, kwargs):
+            for k in kwargs:
+                if k not in self.permitted:
+                    mlog.warning('''Passed invalid keyword argument "{}".'''.format(k))
+                    mlog.warning('This will become a hard error in the future.')
+            return f(obj, args, kwargs)
+        return wrapped
+
+
 class InterpreterException(mesonlib.MesonException):
     pass
 
@@ -89,6 +105,9 @@ class InvalidCode(InterpreterException):
 class InvalidArguments(InterpreterException):
     pass
 
+class SubdirDoneRequest(BaseException):
+    pass
+
 class InterpreterObject:
     def __init__(self):
         self.methods = {}
@@ -187,6 +206,8 @@ class InterpreterBase:
             try:
                 self.current_lineno = cur.lineno
                 self.evaluate_statement(cur)
+            except SubdirDoneRequest:
+                break
             except Exception as e:
                 if not(hasattr(e, 'lineno')):
                     e.lineno = cur.lineno
@@ -408,7 +429,7 @@ The result of this is undefined and will become a hard error in a future Meson r
         varname = node.var_name
         addition = self.evaluate_statement(node.value)
         if is_disabler(addition):
-            set_variable(varname, addition)
+            self.set_variable(varname, addition)
             return
         # Remember that all variables are immutable. We must always create a
         # full new variable and then assign it.
