@@ -21,10 +21,7 @@ from io import StringIO
 from ast import literal_eval
 from enum import Enum
 import tempfile
-from mesonbuild import mtest
-from mesonbuild import environment
-from mesonbuild import mesonlib
-from mesonbuild import mlog
+from mesonbuild import build, environment, mesonlib, mlog, mtest
 from mesonbuild.mesonlib import stringlistify, Popen_safe
 from mesonbuild.coredata import backendlist
 import argparse
@@ -109,8 +106,6 @@ def setup_commands(optbackend):
     if backend is None:
         if msbuild_exe is not None:
             backend = 'vs' # Meson will auto-detect VS version to use
-        elif mesonlib.is_osx():
-            backend = 'xcode'
         else:
             backend = 'ninja'
     # Set backend arguments for Meson
@@ -139,9 +134,9 @@ def get_relative_files_list_from_dir(fromdir):
             paths.append(path)
     return paths
 
-def platform_fix_name(fname, compiler):
+def platform_fix_name(fname, compiler, env):
     if '?lib' in fname:
-        if mesonlib.is_cygwin():
+        if mesonlib.for_cygwin(env.is_cross_build(), env):
             fname = re.sub(r'lib/\?lib(.*)\.so$', r'bin/cyg\1.dll', fname)
             fname = re.sub(r'\?lib(.*)\.dll$', r'cyg\1.dll', fname)
         else:
@@ -149,7 +144,7 @@ def platform_fix_name(fname, compiler):
 
     if fname.endswith('?exe'):
         fname = fname[:-4]
-        if mesonlib.is_windows() or mesonlib.is_cygwin():
+        if mesonlib.for_windows(env.is_cross_build(), env) or mesonlib.for_cygwin(env.is_cross_build(), env):
             return fname + '.exe'
 
     if fname.startswith('?msvc:'):
@@ -164,7 +159,7 @@ def platform_fix_name(fname, compiler):
 
     return fname
 
-def validate_install(srcdir, installdir, compiler):
+def validate_install(srcdir, installdir, compiler, env):
     # List of installed files
     info_file = os.path.join(srcdir, 'installed_files.txt')
     # If this exists, the test does not install any other files
@@ -177,7 +172,7 @@ def validate_install(srcdir, installdir, compiler):
     elif os.path.exists(info_file):
         with open(info_file) as f:
             for line in f:
-                line = platform_fix_name(line.strip(), compiler)
+                line = platform_fix_name(line.strip(), compiler, env)
                 if line:
                     expected[line] = False
     # Check if expected files were found
@@ -209,8 +204,16 @@ def log_text_file(logfile, testdir, stdo, stde):
     logfile.write(stde)
     logfile.write('\n\n---\n\n')
     if print_debug:
-        print(stdo)
-        print(stde, file=sys.stderr)
+        try:
+            print(stdo)
+        except UnicodeError:
+            sanitized_out = stdo.encode('ascii', errors='replace').decode()
+            print(sanitized_out)
+        try:
+            print(stde, file=sys.stderr)
+        except UnicodeError:
+            sanitized_err = stde.encode('ascii', errors='replace').decode()
+            print(sanitized_err, file=sys.stderr)
     if stop:
         print("Aborting..")
         for f in futures:
@@ -316,7 +319,7 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, compiler, backen
     (returncode, stdo, stde) = run_configure(meson_command, gen_args)
     try:
         logfile = os.path.join(test_build_dir, 'meson-logs/meson-log.txt')
-        with open(logfile, errors='ignore') as f:
+        with open(logfile, encoding='utf-8', errors='ignore') as f:
             mesonlog = f.read()
     except Exception:
         mesonlog = no_meson_log_msg
@@ -330,6 +333,7 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, compiler, backen
             return TestResult('Test that should have failed succeeded', BuildStep.configure, stdo, stde, mesonlog, gen_time)
     if returncode != 0:
         return TestResult('Generating the build system failed.', BuildStep.configure, stdo, stde, mesonlog, gen_time)
+    builddata = build.load(test_build_dir)
     # Touch the meson.build file to force a regenerate so we can test that
     # regeneration works before a build is run.
     ensure_backend_detects_changes(backend)
@@ -383,7 +387,8 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, compiler, backen
         return TestResult('Running clean failed.', BuildStep.clean, stdo, stde, mesonlog, gen_time, build_time, test_time)
     if not install_commands:
         return TestResult('', BuildStep.install, '', '', mesonlog, gen_time, build_time, test_time)
-    return TestResult(validate_install(testdir, install_dir, compiler), BuildStep.validate, stdo, stde, mesonlog, gen_time, build_time, test_time)
+    return TestResult(validate_install(testdir, install_dir, compiler, builddata.environment),
+                      BuildStep.validate, stdo, stde, mesonlog, gen_time, build_time, test_time)
 
 def gather_tests(testdir):
     tests = [t.replace('\\', '/').split('/', 2)[2] for t in glob(testdir + '/*')]
@@ -518,7 +523,7 @@ def detect_tests_to_run():
 def run_tests(all_tests, log_name_base, extra_args):
     global logfile
     txtname = log_name_base + '.txt'
-    with open(txtname, 'w', encoding="utf_8") as lf:
+    with open(txtname, 'w', encoding='utf-8', errors='ignore') as lf:
         logfile = lf
         return _run_tests(all_tests, log_name_base, extra_args)
 
@@ -696,7 +701,10 @@ if __name__ == '__main__':
     if failing_tests > 0:
         print('\nMesonlogs of failing tests\n')
         for l in failing_logs:
-            print(l, '\n')
+            try:
+                print(l, '\n')
+            except UnicodeError:
+                print(l.encode('ascii', errors='replace').decode(), '\n')
     for name, dirs, skip in all_tests:
         dirs = (os.path.basename(x) for x in dirs)
         for k, g in itertools.groupby(dirs, key=lambda x: x.split()[0]):
